# Autogenerated by nbdev

d = { 'settings': { 'branch': 'master',
                'doc_baseurl': '/onprem',
                'doc_host': 'https://amaiya.github.io',
                'git_url': 'https://github.com/amaiya/onprem',
                'lib_path': 'onprem'},
  'syms': { 'onprem.console': {},
            'onprem.core': { 'onprem.core.AnswerConversationBufferMemory': ('core.html#answerconversationbuffermemory', 'onprem/core.py'),
                             'onprem.core.AnswerConversationBufferMemory.save_context': ( 'core.html#answerconversationbuffermemory.save_context',
                                                                                          'onprem/core.py'),
                             'onprem.core.LLM': ('core.html#llm', 'onprem/core.py'),
                             'onprem.core.LLM.__init__': ('core.html#llm.__init__', 'onprem/core.py'),
                             'onprem.core.LLM.ask': ('core.html#llm.ask', 'onprem/core.py'),
                             'onprem.core.LLM.chat': ('core.html#llm.chat', 'onprem/core.py'),
                             'onprem.core.LLM.check_model': ('core.html#llm.check_model', 'onprem/core.py'),
                             'onprem.core.LLM.download_model': ('core.html#llm.download_model', 'onprem/core.py'),
                             'onprem.core.LLM.ingest': ('core.html#llm.ingest', 'onprem/core.py'),
                             'onprem.core.LLM.is_azure': ('core.html#llm.is_azure', 'onprem/core.py'),
                             'onprem.core.LLM.is_hf': ('core.html#llm.is_hf', 'onprem/core.py'),
                             'onprem.core.LLM.is_llamacpp': ('core.html#llm.is_llamacpp', 'onprem/core.py'),
                             'onprem.core.LLM.is_local': ('core.html#llm.is_local', 'onprem/core.py'),
                             'onprem.core.LLM.is_local_api': ('core.html#llm.is_local_api', 'onprem/core.py'),
                             'onprem.core.LLM.is_openai_model': ('core.html#llm.is_openai_model', 'onprem/core.py'),
                             'onprem.core.LLM.load_chatqa': ('core.html#llm.load_chatqa', 'onprem/core.py'),
                             'onprem.core.LLM.load_ingester': ('core.html#llm.load_ingester', 'onprem/core.py'),
                             'onprem.core.LLM.load_llm': ('core.html#llm.load_llm', 'onprem/core.py'),
                             'onprem.core.LLM.load_qa': ('core.html#llm.load_qa', 'onprem/core.py'),
                             'onprem.core.LLM.load_vectordb': ('core.html#llm.load_vectordb', 'onprem/core.py'),
                             'onprem.core.LLM.prompt': ('core.html#llm.prompt', 'onprem/core.py'),
                             'onprem.core.LLM.update_max_tokens': ('core.html#llm.update_max_tokens', 'onprem/core.py'),
                             'onprem.core.LLM.update_stop': ('core.html#llm.update_stop', 'onprem/core.py')},
            'onprem.guider': { 'onprem.guider.Guider': ('guider.html#guider', 'onprem/guider.py'),
                               'onprem.guider.Guider.__init__': ('guider.html#guider.__init__', 'onprem/guider.py'),
                               'onprem.guider.Guider.prompt': ('guider.html#guider.prompt', 'onprem/guider.py')},
            'onprem.hf.hfmodel': { 'onprem.hf.hfmodel.HFModel': ('hf.hfmodels.html#hfmodel', 'onprem/hf/hfmodel.py'),
                                   'onprem.hf.hfmodel.HFModel.__init__': ('hf.hfmodels.html#hfmodel.__init__', 'onprem/hf/hfmodel.py'),
                                   'onprem.hf.hfmodel.HFModel.prepare': ('hf.hfmodels.html#hfmodel.prepare', 'onprem/hf/hfmodel.py'),
                                   'onprem.hf.hfmodel.HFModel.tokenize': ('hf.hfmodels.html#hfmodel.tokenize', 'onprem/hf/hfmodel.py')},
            'onprem.hf.hfpipeline': { 'onprem.hf.hfpipeline.HFPipeline': ('hf.hfpipeline.html#hfpipeline', 'onprem/hf/hfpipeline.py'),
                                      'onprem.hf.hfpipeline.HFPipeline.__init__': ( 'hf.hfpipeline.html#hfpipeline.__init__',
                                                                                    'onprem/hf/hfpipeline.py'),
                                      'onprem.hf.hfpipeline.HFPipeline.maxlength': ( 'hf.hfpipeline.html#hfpipeline.maxlength',
                                                                                     'onprem/hf/hfpipeline.py'),
                                      'onprem.hf.hfpipeline.HFPipeline.parseargs': ( 'hf.hfpipeline.html#hfpipeline.parseargs',
                                                                                     'onprem/hf/hfpipeline.py')},
            'onprem.hf.llm.factory': { 'onprem.hf.llm.factory.GenerationFactory': ( 'hf.llm.factory.html#generationfactory',
                                                                                    'onprem/hf/llm/factory.py'),
                                       'onprem.hf.llm.factory.GenerationFactory.create': ( 'hf.llm.factory.html#generationfactory.create',
                                                                                           'onprem/hf/llm/factory.py'),
                                       'onprem.hf.llm.factory.GenerationFactory.method': ( 'hf.llm.factory.html#generationfactory.method',
                                                                                           'onprem/hf/llm/factory.py'),
                                       'onprem.hf.llm.factory.GenerationFactory.resolve': ( 'hf.llm.factory.html#generationfactory.resolve',
                                                                                            'onprem/hf/llm/factory.py')},
            'onprem.hf.llm.generation': { 'onprem.hf.llm.generation.Generation': ( 'hf.llm.generation.html#generation',
                                                                                   'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.__call__': ( 'hf.llm.generation.html#generation.__call__',
                                                                                            'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.__init__': ( 'hf.llm.generation.html#generation.__init__',
                                                                                            'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.clean': ( 'hf.llm.generation.html#generation.clean',
                                                                                         'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.execute': ( 'hf.llm.generation.html#generation.execute',
                                                                                           'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.response': ( 'hf.llm.generation.html#generation.response',
                                                                                            'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.Generation.stream': ( 'hf.llm.generation.html#generation.stream',
                                                                                          'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.TemplateFormatter': ( 'hf.llm.generation.html#templateformatter',
                                                                                          'onprem/hf/llm/generation.py'),
                                          'onprem.hf.llm.generation.TemplateFormatter.check_unused_args': ( 'hf.llm.generation.html#templateformatter.check_unused_args',
                                                                                                            'onprem/hf/llm/generation.py')},
            'onprem.hf.llm.huggingface': { 'onprem.hf.llm.huggingface.Generator': ( 'hf.llm.huggingface.html#generator',
                                                                                    'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.Generator.__init__': ( 'hf.llm.huggingface.html#generator.__init__',
                                                                                             'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFGeneration': ( 'hf.llm.huggingface.html#hfgeneration',
                                                                                       'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFGeneration.__init__': ( 'hf.llm.huggingface.html#hfgeneration.__init__',
                                                                                                'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFGeneration.stream': ( 'hf.llm.huggingface.html#hfgeneration.stream',
                                                                                              'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM': ( 'hf.llm.huggingface.html#hfllm',
                                                                                'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.__call__': ( 'hf.llm.huggingface.html#hfllm.__call__',
                                                                                         'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.__init__': ( 'hf.llm.huggingface.html#hfllm.__init__',
                                                                                         'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.extract': ( 'hf.llm.huggingface.html#hfllm.extract',
                                                                                        'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.parameters': ( 'hf.llm.huggingface.html#hfllm.parameters',
                                                                                           'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.HFLLM.task': ( 'hf.llm.huggingface.html#hfllm.task',
                                                                                     'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.Sequences': ( 'hf.llm.huggingface.html#sequences',
                                                                                    'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.Sequences.__init__': ( 'hf.llm.huggingface.html#sequences.__init__',
                                                                                             'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.StreamingResponse': ( 'hf.llm.huggingface.html#streamingresponse',
                                                                                            'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.StreamingResponse.__call__': ( 'hf.llm.huggingface.html#streamingresponse.__call__',
                                                                                                     'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.StreamingResponse.__init__': ( 'hf.llm.huggingface.html#streamingresponse.__init__',
                                                                                                     'onprem/hf/llm/huggingface.py'),
                                           'onprem.hf.llm.huggingface.StreamingResponse.__iter__': ( 'hf.llm.huggingface.html#streamingresponse.__iter__',
                                                                                                     'onprem/hf/llm/huggingface.py')},
            'onprem.hf.llm.llm': { 'onprem.hf.llm.llm.LLM': ('hf.llm.llm.html#llm', 'onprem/hf/llm/llm.py'),
                                   'onprem.hf.llm.llm.LLM.__call__': ('hf.llm.llm.html#llm.__call__', 'onprem/hf/llm/llm.py'),
                                   'onprem.hf.llm.llm.LLM.__init__': ('hf.llm.llm.html#llm.__init__', 'onprem/hf/llm/llm.py')},
            'onprem.hf.models': { 'onprem.hf.models.Models': ('hf.models.html#models', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.checklength': ('hf.models.html#models.checklength', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.device': ('hf.models.html#models.device', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.deviceid': ('hf.models.html#models.deviceid', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.finddevice': ('hf.models.html#models.finddevice', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.hasaccelerator': ('hf.models.html#models.hasaccelerator', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.hasmpsdevice': ('hf.models.html#models.hasmpsdevice', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.load': ('hf.models.html#models.load', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.maxlength': ('hf.models.html#models.maxlength', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.reference': ('hf.models.html#models.reference', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.task': ('hf.models.html#models.task', 'onprem/hf/models.py'),
                                  'onprem.hf.models.Models.tokenizer': ('hf.models.html#models.tokenizer', 'onprem/hf/models.py')},
            'onprem.hf.onnx': { 'onprem.hf.onnx.OnnxConfig': ('hf.onnx.html#onnxconfig', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel': ('hf.onnx.html#onnxmodel', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.__init__': ('hf.onnx.html#onnxmodel.__init__', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.device': ('hf.onnx.html#onnxmodel.device', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.forward': ('hf.onnx.html#onnxmodel.forward', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.parse': ('hf.onnx.html#onnxmodel.parse', 'onprem/hf/onnx.py'),
                                'onprem.hf.onnx.OnnxModel.providers': ('hf.onnx.html#onnxmodel.providers', 'onprem/hf/onnx.py')},
            'onprem.hf.registry': { 'onprem.hf.registry.Registry': ('hf.registry.html#registry', 'onprem/hf/registry.py'),
                                    'onprem.hf.registry.Registry.register': ( 'hf.registry.html#registry.register',
                                                                              'onprem/hf/registry.py')},
            'onprem.hf.resolver': { 'onprem.hf.resolver.Resolver': ('hf.resolver.html#resolver', 'onprem/hf/resolver.py'),
                                    'onprem.hf.resolver.Resolver.__call__': ( 'hf.resolver.html#resolver.__call__',
                                                                              'onprem/hf/resolver.py')},
            'onprem.hf.tensors': { 'onprem.hf.tensors.Pipeline': ('hf.tensors.html#pipeline', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Pipeline.batch': ('hf.tensors.html#pipeline.batch', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors': ('hf.tensors.html#tensors', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.context': ('hf.tensors.html#tensors.context', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.quantize': ('hf.tensors.html#tensors.quantize', 'onprem/hf/tensors.py'),
                                   'onprem.hf.tensors.Tensors.tensor': ('hf.tensors.html#tensors.tensor', 'onprem/hf/tensors.py')},
            'onprem.ingest': { 'onprem.ingest.Ingester': ('ingest.html#ingester', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.__init__': ('ingest.html#ingester.__init__', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.get_db': ('ingest.html#ingester.get_db', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.get_embedding_model': ( 'ingest.html#ingester.get_embedding_model',
                                                                               'onprem/ingest.py'),
                               'onprem.ingest.Ingester.get_ingested_files': ('ingest.html#ingester.get_ingested_files', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.ingest': ('ingest.html#ingester.ingest', 'onprem/ingest.py'),
                               'onprem.ingest.Ingester.store_documents': ('ingest.html#ingester.store_documents', 'onprem/ingest.py'),
                               'onprem.ingest.MyElmLoader': ('ingest.html#myelmloader', 'onprem/ingest.py'),
                               'onprem.ingest.MyElmLoader.load': ('ingest.html#myelmloader.load', 'onprem/ingest.py'),
                               'onprem.ingest.MyPDFLoader': ('ingest.html#mypdfloader', 'onprem/ingest.py'),
                               'onprem.ingest.MyPDFLoader.load': ('ingest.html#mypdfloader.load', 'onprem/ingest.py'),
                               'onprem.ingest.batchify_chunks': ('ingest.html#batchify_chunks', 'onprem/ingest.py'),
                               'onprem.ingest.does_vectorstore_exist': ('ingest.html#does_vectorstore_exist', 'onprem/ingest.py'),
                               'onprem.ingest.extract_files': ('ingest.html#extract_files', 'onprem/ingest.py'),
                               'onprem.ingest.load_documents': ('ingest.html#load_documents', 'onprem/ingest.py'),
                               'onprem.ingest.load_single_document': ('ingest.html#load_single_document', 'onprem/ingest.py'),
                               'onprem.ingest.process_documents': ('ingest.html#process_documents', 'onprem/ingest.py')},
            'onprem.pipelines.classifier': { 'onprem.pipelines.classifier.ClassifierBase': ( 'pipelines.classifier.html#classifierbase',
                                                                                             'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.arrays2dataset': ( 'pipelines.classifier.html#classifierbase.arrays2dataset',
                                                                                                            'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.dataset2arrays': ( 'pipelines.classifier.html#classifierbase.dataset2arrays',
                                                                                                            'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.evaluate': ( 'pipelines.classifier.html#classifierbase.evaluate',
                                                                                                      'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.explain': ( 'pipelines.classifier.html#classifierbase.explain',
                                                                                                     'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.get_labels': ( 'pipelines.classifier.html#classifierbase.get_labels',
                                                                                                        'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.get_trainer': ( 'pipelines.classifier.html#classifierbase.get_trainer',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.sample_examples': ( 'pipelines.classifier.html#classifierbase.sample_examples',
                                                                                                             'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.save': ( 'pipelines.classifier.html#classifierbase.save',
                                                                                                  'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.ClassifierBase.train': ( 'pipelines.classifier.html#classifierbase.train',
                                                                                                   'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier': ( 'pipelines.classifier.html#fewshotclassifier',
                                                                                                'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.__init__': ( 'pipelines.classifier.html#fewshotclassifier.__init__',
                                                                                                         'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.save': ( 'pipelines.classifier.html#fewshotclassifier.save',
                                                                                                     'onprem/pipelines/classifier.py'),
                                             'onprem.pipelines.classifier.FewShotClassifier.train': ( 'pipelines.classifier.html#fewshotclassifier.train',
                                                                                                      'onprem/pipelines/classifier.py')},
            'onprem.pipelines.extractor': { 'onprem.pipelines.extractor.Extractor': ( 'pipelines.extractor.html#extractor',
                                                                                      'onprem/pipelines/extractor.py'),
                                            'onprem.pipelines.extractor.Extractor.__init__': ( 'pipelines.extractor.html#extractor.__init__',
                                                                                               'onprem/pipelines/extractor.py'),
                                            'onprem.pipelines.extractor.Extractor.apply': ( 'pipelines.extractor.html#extractor.apply',
                                                                                            'onprem/pipelines/extractor.py')},
            'onprem.pipelines.summarizer': { 'onprem.pipelines.summarizer.Summarizer': ( 'pipelines.summarizer.html#summarizer',
                                                                                         'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.__init__': ( 'pipelines.summarizer.html#summarizer.__init__',
                                                                                                  'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer._map_reduce': ( 'pipelines.summarizer.html#summarizer._map_reduce',
                                                                                                     'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer._refine': ( 'pipelines.summarizer.html#summarizer._refine',
                                                                                                 'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.summarize': ( 'pipelines.summarizer.html#summarizer.summarize',
                                                                                                   'onprem/pipelines/summarizer.py'),
                                             'onprem.pipelines.summarizer.Summarizer.summarize_by_concept': ( 'pipelines.summarizer.html#summarizer.summarize_by_concept',
                                                                                                              'onprem/pipelines/summarizer.py')},
            'onprem.utils': { 'onprem.utils.df_to_md': ('utils.html#df_to_md', 'onprem/utils.py'),
                              'onprem.utils.download': ('utils.html#download', 'onprem/utils.py'),
                              'onprem.utils.get_datadir': ('utils.html#get_datadir', 'onprem/utils.py'),
                              'onprem.utils.html_to_df': ('utils.html#html_to_df', 'onprem/utils.py'),
                              'onprem.utils.md_to_df': ('utils.html#md_to_df', 'onprem/utils.py'),
                              'onprem.utils.segment': ('utils.html#segment', 'onprem/utils.py'),
                              'onprem.utils.split_list': ('utils.html#split_list', 'onprem/utils.py')},
            'onprem.webapp': {}}}
